import re

from . import enums
from .misc import Dotdict


NODEINFO_NS = {
	'20': 'http://nodeinfo.diaspora.software/ns/schema/2.0',
	'21': 'http://nodeinfo.diaspora.software/ns/schema/2.1'
}


class HostMeta(str):
	'An object that represents the "/.well-known/host-meta" endpoint'

	def __new__(cls, data):
		return str.__new__(cls, data)


	@classmethod
	def new(cls, domain) -> 'aputils.HostMeta':
		'''
			Generate a new host-meta xml string

			:param str domain: Domain to use for the template url
		'''
		return cls(f'''<?xml version="1.0" encoding="UTF-8">
<XRD xmlns="http://docs.oasis-open.org/ns/xri/xrd-1.0">
	<Link rel="lrdd" template="https://{domain}/.well-known/webfinger?resource={{uri}}" />
</XRD>''')


class HostMetaJson(Dotdict):
	'An object that represents the "/.well-known/host-meta.json" endpoint'

	@classmethod
	def new(cls, domain) -> 'aputils.HostMetaJson':
		'''
			Generate a new host-meta dict

			:param str domain: Domain to use for the template url
		'''
		return cls({
			'links': [
				{
					'rel': 'lrdd',
					'template': f'https://{domain}/.well-known/webfinger?resource={{uri}}'
				}
			]
		})


class Nodeinfo(Dotdict):
	'An object that represents a nodeinfo endpoint'

	@classmethod
	def new(cls, name, version, protocols=[], insrv=[], outsrv=[], metadata={}, **kwargs):
		'''
			Create a new nodeinfo object. It will default to version 2.0 if ``repo`` and ``homeage`` are not set.

			:param str name: Software name to use. Can only include lowercase letters and '-', '_' characters.
			:param str version: Version of the software.
			:param typing.Optional[list] protocols: List of supported protocols (see :py:class:`aputils.NodeinfoProtocol`)
			:param typing.Optional[list] insrv: Supported inbound services (see :py:class:`aputils.NodeinfoServiceInbound`)
			:param typing.Optional[list] outsrv: Supported outbound services (see :py:class:`aputils.NodeinfoServiceOutbound`)
			:param typing.Optional[dict] metadata: Extra server info
			:param typing.Optional[str] repo: Url to the repository that hosts the server code
			:param typing.Optional[str] homepage: Url to the homepage of the server software
			:param typing.Optional[bool] open_regs: Whether user registrations are open or not
			:param typing.Optional[int] users: Total number of registered users
			:param typing.Optional[int] halfyear: Number of active users in the past 6 months
			:param typing.Optional[int] month: Number of active users in the past month
			:param typing.Optional[int] posts: Total number of posts
			:param typing.Optional[int] comments: Total number of comments
		'''

		repo = kwargs.get('repo', None)
		homepage = kwargs.get('homepage', None)
		open_regs = kwargs.get('open_regs', True)
		total_users = kwargs.get('users', 0)
		halfyear_users = kwargs.get('halfyear', 0)
		month_users = kwargs.get('month', 0)
		posts = kwargs.get('posts', 0)
		comments = kwargs.get('comments', 0)

		if not re.match('^[a-z0-9-]+$', name):
			raise ValueError(f'Invalid software name. Must match regex: ^[a-z0-9-]+$')

		if isinstance(protocols, str):
			protocols = [protocols]

		elif not isinstance(protocols, (list, set, tuple)):
			raise TypeError('Protocols must be a list, set, or tuple')

		data = {
			'version': '2.1' if repo or homepage else '2.0',
			'software': {
				'name': name,
				'version': version
			},
			'protocols': list(enums.NodeinfoProtocol.parse(v) for v in protocols),
			'services': {
				'inbound': list(enums.NodeinfoServiceInbound.parse(v) for v in insrv),
				'outbound': list(enums.NodeinfoServiceOutbound.parse(v) for v in outsrv)
			},
			'openRegistrations': open_regs,
			'usage': {
				'localPosts': posts,
				'localComments': comments,
				'users': {
					'total': total_users,
					'activeHalfyear': halfyear_users,
					'activeMonth': month_users,
				}
			},
			'metadata': metadata
		}

		if repo:
			data['software']['repository'] = repo

		if homepage:
			data['software']['homepage'] = homepage

		return cls(data)


	@property
	def sw_name(self) -> str:
		'Get the software name at ``Nodeinfo.software.name``'
		return self.software.name


	@property
	def sw_version(self) -> str:
		'Get the software version at ``Nodeinfo.software.version``'
		return self.software.version


class Webfinger(Dotdict):
	'An object that represents the "/.well-known/webfinger" endpoint'


	@classmethod
	def new(cls, handle, domain, actor, profile=None, interaction=None) -> 'aputils.Webfinger':
		'''
			Create a new webfinger object

			:param str handle: Username of the account
			:param str domain: Domain the account resides on
			:param str actor: URL that points to the ActivityPub Actor of the account
			:param Optional[str] profile: URL that points to an HTML representation of the account
			:param Optional[str] interaction: URL used for remote interaction
		'''
		data = {
			'subject': f'acct:{handle}@{domain}',
			'aliases': [actor],
			'links': [{
				'rel': 'self',
				'type': 'application/activity+json',
				'href': actor
			}]
		}

		if profile:
			data['aliases'].append(profile)
			data['links'].append({
				'rel': 'http://webfinger.net/rel/profile-page',
				'type': 'text/html',
				'href': profile
			})

		if interaction:
			data['links'].append({
				'rel': 'http://ostatus.org/schema/1.0/subscribe',
				'template': interaction
			})

		return cls(data)


	@property
	def handle(self) -> str:
		'Username of the account'
		return self.subject[5:].split('@', 1)[0]


	@property
	def domain(self) -> str:
		'Domain the account resides on'
		return self.subject[5:].split('@', 1)[1]


	@property
	def profile(self) -> str:
		'URL to the HTML representation of the account'
		return self.get_link('profile')


	@property
	def actor(self) -> str:
		'URL that points to the ActivityPub Actor of the account'
		return self.get_link('actor')


	@property
	def interaction(self) -> str:
		'URL used for remote interaction'
		return self.get_link('interaction')


	def get_link(self, type) -> str:
		'''
			Get a URL for the account

			:param str type: The type of URL to get [profile, actor, interaction]
		'''
		if type not in {'profile', 'actor', 'interaction'}:
			raise KeyError(type)

		for link in self.links:
			if type == 'profile' and link.rel == 'http://webfinger.net/rel/profile-page':
				return link.href

			if type == 'actor' and link.rel == 'self':
				return link.href

			if type == 'interaction' and link.rel == 'http://ostatus.org/schema/1.0/subscribe':
				return link.template


class WellKnownNodeinfo(Dotdict):
	'An object that represents the "/.well-known/nodeinfo" endpoint'

	@classmethod
	def new(cls, v20=None, v21=None):
		'''
			Create a new ``WellKnownNodeinfo`` object by specifying the url(s)

			:param typing.Optional[str] v20: URL pointing to a nodeinfo v2.0 object
			:param typing.Optional[str] v21: Url pointing to a nodeinfo v2.1 object
		'''

		if not (v20 or v21):
			raise ValueError('At least one nodeinfo version must be specified')

		data = []

		if v20:
			data.append({'rel': NODEINFO_NS['20'], 'href': v20})

		if v21:
			data.append({'rel': NODEINFO_NS['21'], 'href': v21})

		return cls({'links': data})


	@classmethod
	def new_template(cls, domain, path='/nodeinfo', v20=True, v21=True):
		'''
			Create a new ``WellKnownNodeinfo`` object by specifying the domain, path, and which
			versions to include.

			:param str domain: Domain to use for urls
			:param typing.Optional[str] path: Base path for urls
			:param typing.Optional[bool] v20: If true, generate a Nodeinfo v2.0 url
			:param typing.Optional[bool] v21: If true, generate a Nodeinfo v2.1 url

			:raises ValueError: When both ``v20`` and ``v21`` are ``False``
		'''

		if not (v20 or v21):
			raise ValueError('At least one nodeinfo version must be specified')

		if path.endswith('/'):
			path = path[:-1]

		if not path.startswith('/'):
			path = '/' + path

		urls = {}

		if v20:
			urls['v20'] = f'https://{domain}{path}/2.0.json'

		if v21:
			urls['v21'] = f'https://{domain}{path}/2.1.json'

		return cls.new(**urls)


	@property
	def v20(self) -> str:
		'Try to get the v2.0 url. Shortcut for ``Nodeinfo.get_url("20")``'
		return self.get_url('20')


	@property
	def v21(self) -> str:
		'Try to get the v2.0 url. Shortcut for ``Nodeinfo.get_url("21")``'
		return self.get_url('21')


	def get_url(self, version='20') -> str:
		'''
			Get a versioned nodeinfo url

			:param typing.Optional[str] version: The nodeinfo version url to get

			:raises KeyError: If the url for the specified nodeinfo version cannot be found
		'''

		for item in self.links:
			if item.rel == NODEINFO_NS[version]:
				return item.href

		raise KeyError(version)
